module Emitter where

import Parser
import Table
import Instruction

data Environment = Environment (Table, Int, Int) deriving Show

data EmitValue =
    EmitVoidValue |
    EmitIntValueTemporary Int |
    EmitIntVariable Int String |
    EmitIntValueConstant Int |
    EmitConditionIfFalse (EmitValue, String, [(Instruction, String)], Environment) |
    EmitIntValueRegisterA |
    EmitAddressRegisterY |
    EmitNone deriving (Show)

emitError = error . ("emitter error: " ++)

envTable (Environment (table, offset, label)) = table
envOffset (Environment (table, offset, label)) = offset
envLabel (Environment (table, offset, label)) = label
incEnvOffset (Environment (table, offset, label)) = 
    if offset < 0 then
        emitError "memory over"
    else
        (Environment (table, offset - 1, label))
decEnvOffset (Environment (table, offset, label)) = 
    (Environment (table, offset + 1, label))
incEnvLabel (Environment (table, offset, label)) = (Environment (table, offset , label + 1))

emitValueOfTriple (ev, code, env) = ev
codeOfTriple (ev, code, env) = code
envOfTriple (ev, code, env) = env

emit tree = emitProgram tree (Environment (initialTable "undeclared variable ", 0xf, 0))

emitProgram (Program functions) env = emitConcat emitFunction env functions
emitConcat f env l = concat $ map (\x -> f x env) l

emitFunction (Function (name, ftype, arguments, block)) env =
    codeOfTriple (emitBlock block env)

emitBlock (Block (declarations, statements)) env =
    (emitValueOfTriple statementsResult,
     codeOfTriple statementsResult,
     Environment (envTable env, envOffset env, envLabel (envOfTriple statementsResult)))
    where
      newEnv = emitDeclarations declarations env
      statementsResult = emitStatements statements newEnv

emitStatements = emitSequence emitStatement

emitSequence ef (t:ts) env = 
    (if ts == [] then EmitNone else emitValueOfTriple emitResult,
     (codeOfTriple emitResult) ++ codeOfTriple nextEmitResult,
     envOfTriple nextEmitResult)
    where
      emitResult = ef t env
      nextEmitResult = emitSequence ef ts (envOfTriple emitResult)
emitSequence ef [] env = (EmitNone, [], env)

emitStatement tree env = 
    case tree of
      (Substitution ((Variable variable), expression)) ->
          let 
              emitResult = emitExpression expression env
              emitVal = emitValueOfTriple emitResult
              code = codeOfTriple emitResult
              newEnv = envOfTriple emitResult
              variableOffset = offsetTable (envTable env) variable
          in
            (EmitNone,
            code ++ 
            toRegisterA emitVal ++
            [(OperandInt1 "TIY" variableOffset, variable ++ " = ..."),
             (Operand0 "AM", "")],
             newEnv)
      (If (expression, statements)) ->
          let
              condResult = emitExpression expression env
              condVal = emitValueOfTriple condResult
          in
            case condVal of
              EmitConditionIfFalse (_, condLabel, condCode, condEnv) ->
                  let
                      thenResult = emitStatement statements condEnv
                      newEnv = envOfTriple thenResult
                  in                      
                    (EmitNone,
                     condCode ++
                     codeOfTriple thenResult ++
                     [(Label condLabel, "if end")],
                     newEnv)
              _ ->
                  let
                      condCode = codeOfTriple condResult
                      thenResult = emitStatement statements (envOfTriple condResult)
                      newEnv = envOfTriple thenResult
                      labelNo = envLabel newEnv
                      label1 = emitLabel labelNo
                  in
                    (EmitNone,
                     condCode ++ toRegisterA condVal ++
                     [(OperandInt1 "AIA" 0xf, "if judege"),
                      (OperandInt1 "AIA" 1, ""),
                      (OperandString1 "JUMP" label1, "")] ++ 
                     codeOfTriple thenResult ++
                     [(Label label1, "if end")],
                     incEnvLabel newEnv)

      (IfElse (expression, thenStatements, elseStatements)) ->
          let
              condResult = emitExpression expression env
              condVal = emitValueOfTriple condResult
          in
            case condVal of
              EmitConditionIfFalse (_, condLabel, condCode, condEnv) ->
                  let
                      thenResult = emitStatement thenStatements condEnv
                      elseResult = emitStatement elseStatements (envOfTriple thenResult)
                      newEnv = envOfTriple elseResult
                      labelNo = envLabel newEnv
                      label1 = emitLabel labelNo
                  in
                    (EmitNone,
                     condCode ++
                     codeOfTriple thenResult ++
                     [(FlagResetDontCare, ""),
                      (OperandString1 "JUMP" label1, "if else jump")] ++
                     [(Label condLabel, "else")] ++ 
                     codeOfTriple elseResult ++
                     [(Label label1, "if else end")],
                      incEnvLabel newEnv)
              _ ->
                  let
                      condCode = codeOfTriple condResult
                      thenResult = emitStatement thenStatements (envOfTriple condResult)
                      elseResult = emitStatement elseStatements (envOfTriple thenResult)
                      newEnv = envOfTriple elseResult
                      labelNo = envLabel newEnv
                      label1 = emitLabel labelNo
                      label2 = emitLabel (labelNo + 1)
                  in
                    (EmitNone,
                     condCode ++ toRegisterA condVal ++
                     [(OperandInt1 "CIA" 0, "if else judge"),
                      (OperandString1 "JUMP" label1, "")] ++ 
                     codeOfTriple elseResult ++
                     [(FlagResetDontCare, ""),
                      (OperandString1 "JUMP" label2, "if else jump"),
                      (Label label1, "")] ++
                     codeOfTriple thenResult ++
                     [(Label label2, "if else end")],
                     (incEnvLabel . incEnvLabel) newEnv)

      (While (expression, statements)) ->
          let
              condResult = emitExpression expression env
              condVal = emitValueOfTriple condResult
          in
            case condVal of
              EmitConditionIfFalse (_, condLabel, condCode, condEnv) ->
                  let
                      contentResult = emitStatement statements condEnv
                      newEnv = envOfTriple contentResult
                      labelNo = envLabel newEnv
                      label1 = emitLabel labelNo
                  in
                    (EmitNone,
                     [(Label label1, "while start")] ++
                     condCode ++ 
                     codeOfTriple contentResult ++
                     [(FlagResetDontCare, ""),
                      (OperandString1 "JUMP" label1, "while loop jump"),
                      (Label condLabel, "while end")],
                     incEnvLabel newEnv)
              _ ->
                  let
                      condCode = codeOfTriple condResult
                      contentResult = emitStatement statements (envOfTriple condResult)
                      newEnv = envOfTriple contentResult
                      labelNo = envLabel newEnv
                      label1 = emitLabel labelNo
                      label2 = emitLabel (labelNo + 1)
                  in
                    (EmitNone,
                     [(Label label1, "while start")] ++
                     condCode ++ toRegisterA condVal ++
                     [(OperandInt1 "AIA" 0xf, "while judege"),
                      (OperandInt1 "AIA" 1, ""),
                      (OperandString1 "JUMP" label2, "")] ++
                     codeOfTriple contentResult ++
                     [(FlagResetDontCare, ""),
                      (OperandString1 "JUMP" label1, "while loop jump"),
                      (Label label2, "while end")],
                     (incEnvLabel . incEnvLabel) newEnv)

      (InfiniteLoop (statements)) ->
          let
              emitResult = emitStatement statements env
              newEnv = envOfTriple emitResult
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
          in
            (EmitNone,
            [(Label label1, "infinite loop start")] ++
                                               
            codeOfTriple emitResult ++

            [(FlagResetDontCare, ""),
             (OperandString1 "JUMP" label1, "infinite loop jump")],
             incEnvLabel newEnv)

      (StatementExpression expression) -> emitExpression expression env

      (StatementBlock block) -> emitBlock block env

      (StatementEmpty) -> (EmitNone, [], env)

      _ -> emitError ("unkown statement")

emitExpression tree env = 
    case tree of
      (Number i) -> (EmitIntValueConstant i, [], env)
      (VariableValue (Variable s)) -> (EmitIntVariable (offsetTable (envTable env) s) s, [], env)

      PreOperation1 ("+" , expression) -> emitExpression expression env

      PreOperation1 ("-" , (Number i)) -> (EmitIntValueConstant (mod (0xf + 1 - i) 0x10), [], env)

      PreOperation1 (operator , expression)
          | any (==operator) ["++", "--"] ->
              let
                  emitResult = emitExpression expression env
                  newEnv = envOfTriple emitResult
              in
                (EmitIntValueRegisterA,
                 codeOfTriple emitResult ++
                 addressToRegisterY (emitValueOfTriple emitResult) ++
                 [(Operand0 "MA", ""),
                  (OperandInt1 "AIA" (if operator == "++" then 1 else 0xf), "Unary operator '" ++ operator ++ "'"),
                  (Operand0 "AM", "")],
                 newEnv)

      PreOperation1 ("&" , expression) ->
          let
              emitResult = emitExpression expression env
              newEnv = envOfTriple emitResult
          in
            (EmitAddressRegisterY,
             codeOfTriple emitResult ++
             (addressToRegisterY (emitValueOfTriple emitResult)),
             newEnv)
                                          
      PreOperation1 (operator , expression) ->
          let
              emitResult = emitExpression expression env
              newEnv = envOfTriple emitResult
          in
            case operator of
              "-" ->
                  (EmitIntValueRegisterA,
                   codeOfTriple emitResult ++
                   (toRegisterA (emitValueOfTriple emitResult)) ++
                   [(FlagResetRegistersSafe, ""),
                    (OperandString1 "CAL" "CMPL", "Unary operator '" ++ operator ++ "'"),
                    (OperandInt1 "AIA" 1, "")],
                   newEnv)
              _ -> emitError ("unary operator " ++ operator ++ " not supported")
                  
      PostOperation1 (operator , expression)
          | any (==operator) ["++", "--"] ->
              let
                  emitResult = emitExpression expression env
                  newEnv = envOfTriple emitResult
              in
                (EmitIntValueRegisterA,
                 codeOfTriple emitResult ++
                 addressToRegisterY (emitValueOfTriple emitResult) ++
                 [(Operand0 "MA", ""),
                  (OperandInt1 "AIA" (if operator == "++" then 1 else 0xf), "Unary operator '" ++ operator ++ "'"),
                  (Operand0 "AM", ""),
                  (OperandInt1 "AIA" (if operator == "++" then 0xf else 1), "cancel")],
                 newEnv)
      PostOperation1 (operator , expression) -> emitError ("unary operator " ++ operator ++ " not supported")


      Operation2 (operator , expression1, expression2) ->
          let
              emitResult1 = emitExpression expression1 env
              emitResult2 = emitExpression expression2 (incEnvOffset (envOfTriple emitResult1))
              newEnv = decEnvOffset (envOfTriple emitResult2)
              offset = envOffset newEnv

              reverseEmitResult1 = emitExpression expression2 env
              reverseEmitResult2 = emitExpression expression1 (incEnvOffset (envOfTriple reverseEmitResult1))
              reverseNewEnv = decEnvOffset (envOfTriple reverseEmitResult2)
              reverseOffset = envOffset reverseNewEnv
          in
            case operator of
              o | any (o==) ["+", "-", "!="] -> let op = if o == "!=" then "-" else o in
                  (EmitIntValueRegisterA,
                   codeOfTriple emitResult1 ++
                  (toRegisterA (emitValueOfTriple emitResult1)) ++
                  [(OperandInt1 "TIY" offset, ""),
                   (Operand0 "AM", "")] ++
                   codeOfTriple emitResult2 ++
                  (toRegisterA (emitValueOfTriple emitResult2)) ++
                  [(OperandInt1 "TIY" offset, ""),
                   (Operand0 ("M" ++ op), "Operator '" ++ operator ++ "'")],
                   newEnv)


              o | any (o==) [">=", "<=", ">", "<"] ->
                  let
                      emitResultOp1 = (if o == ">=" || o == "<" then emitResult1 else reverseEmitResult1)
                      emitResultOp2 = (if o == ">=" || o == "<" then emitResult2 else reverseEmitResult2)
                      targetNewEnv = (if o == ">=" || o == "<" then newEnv else reverseNewEnv)
                                      
                      labelNo = envLabel targetNewEnv
                      label1 = emitLabel labelNo
                      label2 = emitLabel (labelNo + 1)
                  in
                    (EmitConditionIfFalse(EmitIntValueRegisterA,
                                          label1,
                                          codeOfTriple emitResultOp1 ++
                                          (toRegisterA (emitValueOfTriple emitResultOp1)) ++
                                          [(OperandInt1 "TIY" offset, ""),
                                           (Operand0 "AM", "")] ++
                                          codeOfTriple emitResultOp2 ++
                                          (toRegisterA (emitValueOfTriple emitResultOp2)) ++
                                          [(OperandInt1 "TIY" offset, ""),
                                           (Operand0 ("M-"), "Operator '" ++ operator ++ "'")] ++
                                          (if o == "<" || o == ">" then [(OperandString1 "JUMP" label2, "")]  else []) ++
                                           [(OperandString1 "JUMP" label1, "OOCV")] ++
                                          (if o == "<" || o == ">" then [(Label label2, "")]  else []),
                                          (incEnvLabel.incEnvLabel) targetNewEnv),
                     codeOfTriple emitResultOp1 ++
                     (toRegisterA (emitValueOfTriple emitResultOp1)) ++
                     [(OperandInt1 "TIY" offset, ""),
                      (Operand0 "AM", "")] ++
                     codeOfTriple emitResultOp2 ++
                     (toRegisterA (emitValueOfTriple emitResultOp2)) ++
                     [(OperandInt1 "TIY" offset, ""),
                      (Operand0 "CH", ""),
                      (if o == "<=" || o == ">="
                       then (OperandInt1 "TIA" 0, "set false")
                       else (OperandInt1 "TIA" 1, "set true")
                      ),
                      (Operand0 "CH", ""),
                      (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                      (OperandString1 "JUMP" label1, ""),
                      (if o == "<=" || o == ">="
                       then (OperandInt1 "TIA" 1, "set true")
                       else (OperandInt1 "TIA" 0, "set false")
                      ),
                      (Operand0 "CH", ""),
                      (Label label1, ""),
                      (Operand0 "CH", "")],
                     incEnvLabel targetNewEnv)

              "==" ->
                  (EmitConditionIfFalse(EmitIntValueRegisterA,
                                        emitLabel (envLabel newEnv),
                                        codeOfTriple emitResult1 ++
                                        (toRegisterA (emitValueOfTriple emitResult1)) ++
                                        [(OperandInt1 "TIY" offset, ""),
                                         (Operand0 "AM", "")] ++
                                        codeOfTriple emitResult2 ++
                                        (toRegisterA (emitValueOfTriple emitResult2)) ++
                                        [(OperandInt1 "TIY" offset, ""),
                                         (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                                         (OperandInt1 "CIA" 0, ""),
                                         (OperandString1 "JUMP" (emitLabel (envLabel newEnv)), "OOCV")],
                                        incEnvLabel newEnv),
                   codeOfTriple emitResult1 ++
                   (toRegisterA (emitValueOfTriple emitResult1)) ++
                   [(OperandInt1 "TIY" offset, ""),
                    (Operand0 "AM", "")] ++
                   codeOfTriple emitResult2 ++
                   (toRegisterA (emitValueOfTriple emitResult2)) ++
                   [(OperandInt1 "TIY" offset, ""),
                    (FlagResetRegistersSafe, ""),
                    (OperandString1 "CAL" "CHNG", "set false Ar'"),
                    (OperandInt1 "TIA" 0, ""),
                    (OperandInt1 "TIY" offset, ""),
                    (FlagResetRegistersSafe, ""),
                    (OperandString1 "CAL" "CHNG", ""),
                    (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                    (OperandInt1 "AIA" 0xf, ""),
                    (OperandString1 "CAL" "CHNG", "")],
                   newEnv)

              _ -> emitError ("operator " ++ operator ++ " not supported")

      (ExpressionCall0 fname) -> emitExpressionCall0 fname env
      (ExpressionCall1 (fname, expression)) -> emitExpressionCall1 fname expression env

      _ -> emitError ("unkown expression")



emitDeclarations declarations env = foldl (\x y -> emitDeclaration y x) env declarations

emitDeclaration (Declaration (dtype, variables)) env =
    foldl (\x y -> emitDeclarationVariable dtype y x) env variables

emitDeclarationVariable TypeInt (Variable vname) env =
    if offset < 0 then
        emitError "memory over"
    else
        Environment (newTable, offset - 1, envLabel env)
    where
      offset = envOffset env
      newTable = updateTable (envTable env) vname offset

emitExpressionCall0 fname env = 
    case fname of
      "exit" ->
          let
              labelNo = envLabel env
              label1 = emitLabel labelNo
          in
            (EmitVoidValue,
             [(FlagResetDontCare, ""),
              (Label label1, ""),
              (OperandString1 "JUMP" label1, "exit loop")],
             incEnvLabel env)
          
      _ ->
          (EmitVoidValue, [(FlagResetDontCare, "")] ++ emitFunctionCall0 fname, env)

emitFunctionCall0 fn@"resetNumberLed" = [(OperandString1 "CAL" "RSTO", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepEndSound" = [(OperandString1 "CAL" "ENDS", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepErrorSound" = [(OperandString1 "CAL" "ERRS", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepShortSound" = [(OperandString1 "CAL" "SHTS", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepLongSound" = [(OperandString1 "CAL" "LONS", "call " ++ fn ++ "()")]
emitFunctionCall0 fname = emitError ("unkown function " ++ fname ++ "(void)")

emitExpressionCall1 fname expression env = 
    case fname of
      "inputKey" ->
          let
              emitResult = emitExpression expression env
              newEnv = envOfTriple emitResult
              addressCalcCode = codeOfTriple emitResult
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
          in
            case addressCalcCode of
              [(OperandInt1 "TIY" _, _)] ->

                  (EmitConditionIfFalse(EmitIntValueRegisterA,
                                        label1,
                                        [(Operand0 "KA", ""),
                                         (OperandString1 "JUMP" label1, "OOCV")] ++
                                         addressCalcCode ++
                                         [(Operand0 "AM", "")],
                                         incEnvLabel newEnv),

                   [(OperandInt1 "TIA" 0, "inputKey start"),
                    (Operand0 "CH" , ""),
                    (Operand0 "KA", ""),
                    (OperandString1 "JUMP" label1, "")] ++
                   addressCalcCode ++
                   [(Operand0 "AM", ""),
                    (OperandInt1 "TIA" 1, ""),
                    (Operand0 "CH" , ""),
                    (Label label1, ""),
                    (Operand0 "CH" , "")],
                   incEnvLabel newEnv)

              _ -> emitError ("can't emit expression in inputKey(..), code size too large")
      _ ->
          let
              emitResult = emitExpression expression env
          in
            (EmitVoidValue,
             codeOfTriple emitResult ++
             toRegisterA (emitValueOfTriple emitResult) ++
             emitFunctionCall1 fname,
             envOfTriple emitResult)

emitFunctionCall1 fn@"setNumberLed" = [(Operand0 "AO", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"setLed" = [(Operand0 "CY", ""), (OperandString1 "CAL" "SETR", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"resetLed" = [(Operand0 "CY", ""), (OperandString1 "CAL" "RSTR", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"beep" = [(FlagResetRegistersSafe, "")] ++ [(OperandString1 "CAL" "SUND", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"sleep" = [(FlagResetRegistersSafe, "")] ++ [(OperandString1 "CAL" "TIMR", "call " ++ fn ++ "()")]
emitFunctionCall1 fname = emitError ("unkown function " ++ fname ++ "(int)")

toRegisterA EmitIntValueRegisterA = []
toRegisterA (EmitIntValueTemporary i) =
    [(OperandInt1 "TIY" i, "Temporary offset " ++ show i) , (Operand0 "MA", "")]
toRegisterA (EmitIntVariable i s) =
    [(OperandInt1 "TIY" i, "Variable '" ++ s ++ "' (offset " ++ show i ++ ")") , (Operand0 "MA", "")]
toRegisterA (EmitIntValueConstant i) = 
    [(OperandInt1 "TIA" i, "Constant " ++ show i)]
toRegisterA EmitAddressRegisterY = 
    [(Operand0 "CY", "")]
toRegisterA (EmitConditionIfFalse (ev, _, _, _)) = toRegisterA ev
toRegisterA EmitVoidValue = emitError ("void value not ignored as it ought to be")

addressToRegisterY EmitAddressRegisterY = []
addressToRegisterY (EmitIntVariable i s) =
    [(OperandInt1 "TIY" i, "Variable '" ++ s ++ "' (offset " ++ show i ++ ")")]
addressToRegisterY t = emitError ("invalid reference " ++ show t)

emitLabel = ("LABEL" ++) . show
