module Parser where

import Lexer

data Program = Program [Function] deriving (Show, Eq)
data Function = NoFunction Error | Function (String, Type, Arguments, Block) deriving (Show, Eq)
data Type = NoType Error | TypeInt | TypeVoid deriving (Show, Eq)
data Block = NoBlock Error | Block ([Declaration], [Statement]) deriving (Show, Eq)
data Statement =
    NoStatement Error |
    StatementEmpty |
    Substitution (Variable, Expression) |
    If (Expression, Statement) |
    IfElse (Expression, Statement, Statement) |
    While (Expression, Statement) |
    InfiniteLoop (Statement) |
    StatementCall0 (String) |
    StatementCall1 (String, Expression) |
    StatementBlock Block |
    StatementExpression (Expression)
    deriving (Show, Eq)
data Variable = NoVariable Error | Variable String deriving (Show, Eq)
data Expression =
    Number Int |
    VariableValue Variable |
    PreOperation1 (String, Expression) |
    PostOperation1 (String, Expression) |
    Operation2 (String, Expression, Expression) |
    ExpressionCall0 (String) |
    ExpressionCall1 (String, Expression) |
    NoExpression Error
    deriving (Show, Eq)
data Arguments = Arguments [Argument] | ArgumentsVoid deriving (Show, Eq)
data Argument = NoArgument Error | Argument (Type, Variable) deriving (Show, Eq)
data Declaration = NoDeclaration Error | Declaration (Type, [Variable]) deriving (Show, Eq)

data Error = Error [String] | NoError deriving (Show, Eq)

functionError (NoFunction e) = e
functionError _ = NoError
expressionError (NoExpression e) = e
expressionError _ = NoError
variableError (NoVariable e) = e
variableError _ = NoError
argumentError (NoArgument e) = e
argumentError _ = NoError
declarationError (NoDeclaration e) = e
declarationError _ = NoError
statementError (NoStatement e) = e
statementError _ = NoError
typeError (NoType e) = e
typeError _ = NoError
           
tree = fst
remainTokens = snd

parse tokens =
    if remainTokens program == [] then
        tree program
    else
        parseErrorStop "invalid syntax" (remainTokens program)
    where
      program = parseProgram tokens

parseError t message tokens = (t (Error [parseErrorMessage message tokens]), tokens)
parseErrorStop message tokens = error (parseErrorMessage message tokens)
parseErrorMessage message tokens = ("parse error: " ++ message ++ " at " ++ errorTokenShow tokens)
errorTokenShow (token:tokens) = show token
errorTokenShow _ = "?"

eatCheckToken t (token:tokens) | token == t = tokens
eatCheckToken t tokens = parseErrorStop ("can't found " ++ show t) tokens

eatCheckTokens (t:ts) tokens = eatCheckTokens ts (eatCheckToken t tokens)
eatCheckTokens [] tokens = tokens

parseProgram tokens = 
    (Program (tree functions), (remainTokens functions))
    where
      functions = parseFunctions tokens

parseFunctions = parseSequence parseFunction functionError

parseFunction tokens =
    case ftypeRemain of
      ((TokenIdentifier fname):(TokenSymbol "("):rtokens) ->
          let 
              args = parseArguments rtokens
              argsRemain = eatCheckToken (TokenSymbol ")") (remainTokens args)
              block = parseBlock argsRemain
          in
            (Function (fname, tree ftype, tree args, tree block), remainTokens block)
      _ -> parseError NoFunction "invalid function" tokens
    where
      ftype = parseType tokens
      ftypeRemain = remainTokens ftype

parseBlock ((TokenSymbol "{"):tokens) = 
    (Block (tree declarations, tree statements), eatCheckToken (TokenSymbol "}") (remainTokens statements))
    where
      declarations = parseDeclarations tokens
      statements = parseStatements (remainTokens declarations)
parseBlock tokens = parseError NoBlock "invalid block" tokens

parseStatements = parseSequence parseStatement statementError

parseStatement ((TokenIdentifier "while"):(TokenSymbol "("):tokens) =
    case tree cond of
      (Number 0) -> (StatementEmpty, remainTokens cont)
      (Number i) -> (InfiniteLoop (tree cont), remainTokens cont)
      _ -> (While (tree cond, tree cont), remainTokens cont)
    where
      cond = parseExpression tokens
      cont = parseStatement (eatCheckToken (TokenSymbol ")") (remainTokens cond))
parseStatement ((TokenIdentifier "if"):(TokenSymbol "("):tokens) =
    if head thenRemain == (TokenIdentifier "else") then
        let
            elseStatemant = parseStatement (tail thenRemain)
        in
          (IfElse (tree cond, tree thenStatemant, tree elseStatemant), remainTokens elseStatemant)
    else
        (If (tree cond, tree thenStatemant), thenRemain)
    where
      cond = parseExpression tokens
      thenStatemant = parseStatement (eatCheckToken (TokenSymbol ")") (remainTokens cond))
      thenRemain = (remainTokens thenStatemant)

parseStatement atokens@((TokenSymbol "{"):tokens) =
    (StatementBlock (tree block), remainTokens block)
    where
      block = parseBlock atokens
             
parseStatement atokens@((TokenIdentifier vname):(TokenSymbol "="):tokens) =
    (Substitution (tree variable, tree expression), eatCheckToken (TokenSymbol ";") (remainTokens expression))
    where
      variable = parseVariable atokens
      expression = parseExpression (eatCheckToken (TokenSymbol "=") (remainTokens variable))

parseStatement tokens =
    if (expressionError (tree expression)) == NoError
    then
        (StatementExpression (tree expression), eatCheckToken (TokenSymbol ";") (remainTokens expression))
    else
        parseError NoStatement "invalid statement" tokens
    where
      expression = parseExpression tokens

parseDeclarations = parseSequence parseDeclaration declarationError

parseDeclaration tokens =
    if typeError (tree ts) == NoError && (tree vs) /= [] then
        (Declaration (tree ts, tree vs),
         eatCheckToken (TokenSymbol ";") (remainTokens vs))
    else
        parseError NoDeclaration "invalid declaration" tokens
    where
      ts = parseType tokens
      vs = parseVariables (remainTokens ts)

parseVariables = parseEnumeration (TokenSymbol ",") parseVariable variableError

parseVariable ((TokenIdentifier vname):tokens) = ((Variable vname), tokens)
parseVariable tokens = parseError NoVariable "invalid variable" tokens

parseArguments ((TokenIdentifier "void"):(TokenSymbol ")"):tokens) = 
    (ArgumentsVoid, (TokenSymbol ")"):tokens)
parseArguments tokens =
    (Arguments (tree arguments), remainTokens arguments)
    where
      arguments = parseEnumeration (TokenSymbol ",") parseArgument argumentError tokens

parseArgument tokens =
    if typeError (tree ts) == NoError && variableError (tree vs) == NoError then
        (Argument (tree ts, tree vs), (remainTokens vs))
    else
        parseError NoArgument "invalid argument" tokens
    where
      ts = parseType tokens
      vs = parseVariable (remainTokens ts)

parseExpression = parseOperation2LeftExpression

parseOperation2LeftExpression =
    foldl parseOperation2Left parsePreOperation1RightExpression
              [(\x -> any (==x) ["*", "/"]),
               (\x -> any (==x) ["+", "-"]),
               (\x -> any (==x) ["<", "<=", ">", ">="]),
               (\x -> any (==x) ["==", "!="]),
               (\x -> x == "||"),
               (\x -> x == "&&")]

parsePreOperation1RightExpression =
    foldl parsePreOperation1Right parsePostOperation1LeftExpression
              [(\x -> any (==x) ["++", "--", "+", "-", "&"])]

parsePostOperation1LeftExpression =
    foldl parsePostOperation1Left parseFactor
              [(\x -> any (==x) ["++", "--"])]

parseFactor ((TokenNumber i):tokens) = (Number i, tokens)
parseFactor ((TokenIdentifier fname):(TokenSymbol "("):(TokenSymbol ")"):tokens) =
    (ExpressionCall0 fname, tokens);
parseFactor ((TokenIdentifier fname):(TokenSymbol "("):tokens) = 
    (ExpressionCall1 (fname, tree expression), eatCheckToken (TokenSymbol ")") (remainTokens expression))
    where
      expression = parseExpression tokens

parseFactor ((TokenIdentifier v):tokens) = (VariableValue (Variable v), tokens)
parseFactor ((TokenSymbol "("):tokens) =
    if head r /= (TokenSymbol ")") then
        parseError NoExpression "invalid factor" r
    else
        (tree t, tail r)
    where
      t = parseExpression tokens
      r = remainTokens t
parseFactor tokens = parseError NoExpression "invalid factor" tokens

parseType (TokenIdentifier "int":tokens) = (TypeInt, tokens)
parseType (TokenIdentifier "void":tokens) = (TypeVoid, tokens)
parseType tokens = parseError NoType "invalid type" tokens

parseSequence pf errorGetter tokens =
    if (errorGetter at) == NoError then
        let next = parseSequence pf errorGetter r
            nextTree = tree next
            nextTokens = remainTokens next
        in
          (at:nextTree, nextTokens)
    else
        ([], tokens)
    where
      a = pf tokens
      at = tree a
      r = remainTokens a

parseEnumeration delimiterToken pf errorGetter tokens =
    if (errorGetter at) == NoError then
        case r of
          (t:rr) | delimiterToken == t -> 
               let next = parseEnumeration delimiterToken pf errorGetter rr
                   nextTree = tree next
                   nextTokens = remainTokens next
               in
                 (at:nextTree, nextTokens)
          _ -> ([at], r)
    else
        ([], tokens)
    where
      a = pf tokens
      at = tree a
      r = remainTokens a

parsePostOperation1Left p1 isSymbol tokens =
    parsePostOperation1Left' p1 isSymbol (remainTokens t) (tree t)
    where
      t = p1 tokens
parsePostOperation1Left' p1 isSymbol atokens@((TokenSymbol s):tokens) t =
    if isSymbol s
    then
        parsePostOperation1Left' p1 isSymbol tokens (PostOperation1 (s, t))
    else
        (t, atokens)
parsePostOperation1Left' p1 isSymbol tokens t = (t, tokens)

parsePreOperation1Right p1 isSymbol ((TokenSymbol s):tokens)
    | isSymbol s =
        (PreOperation1 (s, tree t), remainTokens t)
        where
          t = parsePreOperation1Right p1 isSymbol tokens
parsePreOperation1Right p1 isSymbol tokens = p1 tokens

parseOperation2Left p1 isSymbol tokens =
    parseOperation2Left' p1 isSymbol (remainTokens t) (tree t)
    where
      t = p1 tokens
parseOperation2Left' p1 isSymbol atokens@((TokenSymbol s):tokens) t = 
    if isSymbol s then
        parseOperation2Left' p1 isSymbol (remainTokens t1) (Operation2 (s, t, tree t1))
    else (t, atokens)
    where
      t1 = p1 tokens
parseOperation2Left' p1 isSymbol tokens t = (t, tokens)

parseOperation2Right p1 isSymbol tokens =
    parseOperation2Right' p1 isSymbol (remainTokens t) (tree t)
    where
      t = p1 tokens
parseOperation2Right' p1 isSymbol atokens@((TokenSymbol s):tokens) t = 
    if isSymbol s then
        (Operation2 (s, t, tree t1), (remainTokens t1))
    else (t, atokens)
    where t1 = parseOperation2Right p1 isSymbol tokens
parseOperation2Right' p1 isSymbol tokens t = (t, tokens)
